import threading
import multiprocess as mp
import time
import rpyc
import uuid
import math
from rpyc.utils.server import ThreadedServer

@rpyc.service
class CustomService(rpyc.Service):
    def __init__(self, function):
        self.function = function
        self.lock = threading.Lock()
        
    @rpyc.exposed
    def get_data(self):
        data = []
        try:
            self.lock.acquire()
            data = self.function()
            self.lock.release()
        except Exception as ex:
            return 'ERROR'
        return data            
    
    @rpyc.exposed
    def stop_server(self):
        self.server.close()
            
    def update_server(self, server):
        self.server = server

class LivePlotter():  
    def __init__(self, logger, num_samples = 200, data_rate = 100, plot_rate = 10) -> None:
        self.count = 0
        self._num_samples = num_samples
        self._data_rate = data_rate
        self._plot_rate = plot_rate
        self._logger = logger
        
    def plot(self, get_data_callback):

        def plot_data(count, plot_rate, num_samples, data_rate):
            import matplotlib.pyplot as plt
            import numpy as np
            import matplotlib.animation as animation

            cancellation_token = threading.Event()
            conn = rpyc.connect("localhost", port=(18820 + count), config={'allow_public_attrs': True, 'import_custom_exceptions': True})
            
            global vals
            vals = np.ndarray((0,0))
                        
            def closed(evt):
                cancellation_token.set()
                
            def fetch_data():
                global vals
                while not cancellation_token.is_set():
                    data = conn.root.get_data()
                    
                    if(data == 'ERROR'):
                        cancellation_token.set()
                        break   
                    
                    if(len(vals) == 0):
                        vals = data
                    else:
                        vals = np.vstack((vals, data))        
                
                    if(len(vals) > num_samples):
                        vals = vals[-num_samples:]
                    
                    time.sleep(1/data_rate)
                
            fetch_data_f = threading.Thread(target=fetch_data)
                            
            fig, ax = plt.subplots()
            try:
                fig.canvas.set_window_title('Plot ' + str(uuid.uuid4())[:4])
            except:
                self._logger.warn("Unable to set window title, check if matplotlib version is correct as per requirements.txt")
            fig.canvas.mpl_connect('close_event', closed)

            init_data = conn.root.get_data()
            
            if(init_data == 'ERROR'):
                cancellation_token.set()
                self._logger.error("Plot Initialisation Failed. Please check callback and try again")
                return
            
            init_data = np.array(init_data)                       
                                
            if(not (init_data.ndim == 1)):
                cancellation_token.set()
                self._logger.error("Plotter only supports 1D arrays.")
                return     

            lines = []
            for i in range(len(init_data)):
                line, = ax.plot(0,init_data[i], label = f'{i}')
                lines.append(line)

            plt.legend()

            fetch_data_f.start()
            
            time.sleep(0.5)          
            
            def update(frame):
                if(cancellation_token.is_set()):
                    plt.close(fig)
                
                for i in range(len(lines)):
                    lines[i].set_data(np.arange(0,len(vals[:, i])), vals[:, i])
                    # lines[i].set_data(np.arange(0,100), np.random.randn(100))
                
                y_min = np.min(vals) - 0.5*abs(np.min(vals))
                y_max = np.max(vals) + 1.0*abs(np.max(vals))
                
                # print(y_min, y_max, np.min(vals), np.max(vals))
                
                ax.set_ylim([y_min, y_max])
                ax.set_xlim([0, len(vals)])
                
                return tuple(lines)
                        
            plot_anim = animation.FuncAnimation(fig=fig, func=update, interval=plot_rate)
            plt.show()
            
            fetch_data_f.join()
            
            try:    
                conn.root.stop_server()
            except EOFError as e:
                # print("Server was closed")
                pass
            conn.close()
        
        def server_thread():  
            gen_data_service = CustomService(get_data_callback)    
            data_server = ThreadedServer(gen_data_service, port=(18820 + self.count), protocol_config={'allow_public_attrs': True})
            gen_data_service.update_server(data_server)
            self.count = self.count + 1
            data_server.start()
            
        data_server_t = threading.Thread(target=server_thread, daemon=True)
        plot_data_p = mp.Process(target=plot_data, args=(self.count,self._plot_rate, self._num_samples, self._data_rate), daemon=True)
        
        data_server_t.start()
        plot_data_p.start()            
        